-- primary decomposition of an ideal I by SYCI algorithm
-- output : {QL0,QL1,...}
-- QLi={Qi0,Qi1,...} is the list of all primary components of level i.
-- In particular QL0 is the list of all isolated primary components.

sycidec = (I) -> (
  local PLi,QLi,RLi,Si,Ci,Yi,Ti;
  
  -- Ri is the intersection of all primary components of level < i
  R := ring I; Ri := ideal 1_R; QL := {};
  for i from 0  do (
    << "level=" << i << ": " << flush;
    -- compute the minimal associated primes of I:Ri
    PLi = colonMinimalPrimes(I,Ri);
    -- compute a separator set
    Si = separator(I,PLi);
    -- Rij = Ri cap ((sat(I,Sij)R_Yij) cap R)
    Ci = apply(Si,f->saturate(I,f));
    Yi = apply(PLi,P->independentSets(P,Limit=>1));
    RLi = apply(Ci,Yi,(c,y)->intersect(Ri,ideal gens gb myextract(c,y)));
    -- if i==0 then Rij is an isolated primary component
    -- else Rij=Ri cap Qij, so we have to compute 
    -- a saturated separating ideal for (Ri,Rij) to extract Qij.
    if i == 0 then ( Rad := intersect(PLi); QLi = RLi )
    else (
      -- Rij=Ri cap Ti
      -- Qij = (Ti)R_Yij cap R
      Ti = apply(PLi,RLi,(p,r)->r+saturatedSeparatingIdeal(p,r,Ri,Rad));
      QLi = apply(Ti,Yi,(t,y)->ideal gens gb myextract(t,y))
    );
    << endl << flush;
    QL = append(QL,QLi); Ri = intersect(RLi);
    if Ri == I then return QL
  )
)

removeRedundantComps = (L) -> (
  if #L == 1 then return L;
  S := new MutableList from L;
  for i from 0 to #L-1 do (
     if S#i == 0 then continue;
     for j from 0 to #L-1 do (
         if j == i or S#j == 0 then continue;
         if S#i == S#j or isSubset(S#i,S#j) then S#j = 0
       )
  );
  return toList(select(S,i -> i!=0))
)

-- prime decomposition of rad(I:J)

colonMinimalPrimes = (I,J) -> (
   local K,PL,S;

   << "minimalPrimes:" << flush;
   R := ring I; L := {};
   for f in first entries mingens J do (
     if f==1 then K=I else K=I:f;
     if K != ideal 1_R then L = append(L,K)
   );
   << #L << "->" << flush;
   L = removeRedundantComps(L);
   << #L << flush;
   P := {};
   for K in L do (
       <<"." <<flush;
       S = apply(first entries gens K,f->squarefree(f));
       PL = minimalPrimes ideal mingens gb ideal S;
       P = join(PL,P)
   );
   P = removeRedundantComps(P);
   << " comps=" << #P << endl << flush;
   return P
)

mindeg = (L) -> (
  -- select the element of least degree
  f := L#0;
  degf := degree f;
  scan(L, g -> if degree g < degf then (f = g; degf = degree g));
  return f 
)

-- compute (IR_Y cap R) for a pseudo-primary ideal I
-- where Y is a maximally independent set of rad(I)
-- calls flattener in Shimoyama-Yokoyama.m2

load "PrimaryDecomposition/Shimoyama-Yokoyama.m2";

myextract = (I,Y) -> (
  R := ring I;
  if #Y == 0 then f = 1_R
  else f := flattener(I,Y#0);
  if f != 1_R then return saturate(I,f)
  else return I
)


-- compute a saturated separating ideal J for (C,I,Q) 
-- s.t. I=(Q cap (I+J)) and rad(C) = rad(I:Q) = rad(I+J)

saturatedSeparatingIdeal = (C,I,Q,Rad) -> (
   local fi;

   << "." << flush;
   S := nonmember(C,Rad);
   if intersect(I+ideal S,Q) == I then return S;
   I1 := I;
   SSI := {};
   for f in S do (
     fi = f;
     while (intersect(Q,I1+ideal fi) != I) do (fi=fi*f);
        I1 = I1+ideal fi;
     SSI=append(SSI,fi)
   );
   return ideal SSI
)

-- compute f in P1 setminus P2

nonmember = (P1,P2) ->
  select(first entries gens P1, f->not isSubset(ideal f,P2))

squarefree = (f) -> value apply(factor f, i -> i#0)

separator = (I,PP) -> (
  R := ring I;
  S := new MutableList;
  scan(toList(0..#PP-1),i->S#i=1_R);
  scan(toList(0..#PP-1),i->
     scan(toList(0..#PP-1),j->
       if i != j then S#j =  lcm(S#j,squarefree(mindeg nonmember(PP#i,PP#j))))
  );
  return S
)
